from math import erfc
import numpy as np
from scipy.special import erfc


def get_quantiles(acquisition_par, fmin, m, s):
    '''
    Quantiles of the Gaussian distribution useful to determine the acquisition function values
    :param acquisition_par: parameter of the acquisition function
    :param fmin: current minimum.
    :param m: vector of means.
    :param s: vector of standard deviations.
    '''
    if isinstance(s, np.ndarray):
        s[s<1e-10] = 1e-10
    elif s< 1e-10:
        s = 1e-10
    u = (fmin - m - acquisition_par)/s

    phi = np.exp(-0.5 * u**2) / np.sqrt(2*np.pi)
    # vectorized version of erfc to not depend on scipy
    Phi = 0.5 * erfc(-u / np.sqrt(2))
    return (phi, Phi, u)



from abc import ABC, abstractmethod
import numpy as np


class Scaling(ABC):
    @abstractmethod
    def to_internal(self, value: float) -> float:
        pass

    @abstractmethod
    def from_internal(self, value: float) -> float:
        pass

    def __repr__(self):
        return "{}()".format(self.__class__.__name__)

    def __eq__(self, other):
        # For usage in tests. Make sure to edit if parameters are added.
        return self.__class__ == other.__class__


class LinearScaling(Scaling):
    def to_internal(self, value: float) -> float:
        return value

    def from_internal(self, value: float) -> float:
        return value


class LogScaling(Scaling):
    def to_internal(self, value: float) -> float:
        assert value > 0, "Value must be strictly positive to be log-scaled."
        return np.log(value)

    def from_internal(self, value: float) -> float:
        return np.exp(value)


class ReverseLogScaling(Scaling):
    def to_internal(self, value: float) -> float:
        assert 0 <= value < 1, \
            "Value must be between 0 (inclusive) and 1 (exclusive) to be reverse-log-scaled."
        return -np.log(1.0 - value)

    def from_internal(self, value: float) -> float:
        return 1.0 - np.exp(-value)



import pickle
import logging

from ..utils import warning_filter
with warning_filter():
    from skopt import Optimizer
    from skopt.space import Integer, Real, Categorical

from .searcher import BaseSearcher

__all__ = ['SKoptSearcher']

logger = logging.getLogger(__name__)


class SKoptSearcher(BaseSearcher):
    """SKopt Searcher that uses Bayesian optimization to suggest new hyperparameter configurations. 
        Requires that 'scikit-optimize' package is installed.
    
    Parameters
    ----------
    configspace: ConfigSpace.ConfigurationSpace
        The configuration space to sample from. It contains the full
        specification of the Hyperparameters with their priors
    kwargs: Optional arguments passed to skopt.optimizer.Optimizer class. 
        Please see documentation at this link: `skopt.optimizer.Optimizer <http://scikit-optimize.github.io/optimizer/index.html#skopt.optimizer.Optimizer>`_
        These kwargs be used to specify which surrogate model Bayesian optimization should rely on,
        which acquisition function to use, how to optimize the acquisition function, etc.
        The skopt library provides comprehensive Bayesian optimization functionality,
        where popular non-default kwargs options here might include: 
        
        - base_estimator = 'GP' or 'RF' or 'ET' or 'GBRT' (to specify different surrogate models like Gaussian Processes, Random Forests, etc)
        
        - acq_func = 'LCB' or 'EI' or 'PI' or 'gp_hedge' (to specify different acquisition functions like Lower Confidence Bound, Expected Improvement, etc)
        
        For example, we can tell our Searcher to perform Bayesian optimization with a Random Forest surrogate model
        and use the Expected Improvement acquisition function by invoking: 
        `SKoptSearcher(cs, base_estimator='RF', acq_func='EI')`
    
    Examples
    --------
    By default, the searcher is created along with the scheduler. For example:

    >>> import autogluon as ag
    >>> @ag.args(
    ...     lr=ag.space.Real(1e-3, 1e-2, log=True))
    >>> def train_fn(args, reporter):
    ...     reporter(accuracy = args.lr ** 2)
    >>> search_options = {'base_estimator': 'RF', 'acq_func': 'EI'}
    >>> scheduler = ag.scheduler.FIFOScheduler(
    ...     train_fn, searcher='skopt', search_options=search_options,
    ...     num_trials=10, reward_attr='accuracy')

    This would result in a SKoptSearcher with cs = train_fn.cs. You can also
    create a SKoptSearcher by hand:

    >>> import autogluon as ag
    >>> @ag.args(
    ...     lr=ag.space.Real(1e-3, 1e-2, log=True),
    ...     wd=ag.space.Real(1e-3, 1e-2))
    >>> def train_fn(args, reporter):
    ...     pass
    >>> searcher = ag.searcher.SKoptSearcher(train_fn.cs)
    >>> searcher.get_config()
    {'lr': 0.0031622777, 'wd': 0.0055}
    >>> searcher = SKoptSearcher(
    ...     train_fn.cs, reward_attribute='accuracy', base_estimator='RF',
    ...     acq_func='EI')
    >>> next_config = searcher.get_config()
    >>> searcher.update(next_config, accuracy=10.0)  # made-up value
    
    .. note::
        
        - get_config() cannot ensure valid configurations for conditional spaces since skopt 
        does not contain this functionality as it is not integrated with ConfigSpace. 
        If invalid config is produced, `SKoptSearcher.get_config()` will catch these Exceptions and revert to `random_config()` instead.
        
        - get_config(max_tries) uses skopt's batch BayesOpt functionality to query at most 
        max_tries number of configs to try out.
        If all of these have configs have already been scheduled to try (might happen in asynchronous setting), 
        then get_config simply reverts to random search via random_config().
    """
    errors_tohandle = (ValueError, TypeError, RuntimeError)

    def __init__(self, configspace, **kwargs):
        super().__init__(
            configspace, reward_attribute=kwargs.get('reward_attribute'))
        self.hp_ordering = configspace.get_hyperparameter_names() # fix order of hyperparams in configspace.
        skopt_hpspace = []
        for hp in self.hp_ordering:
            hp_obj = configspace.get_hyperparameter(hp)
            hp_type = str(type(hp_obj)).lower() # type of hyperparam
            if 'integer' in hp_type:
                hp_dimension = Integer(low=int(hp_obj.lower), high=int(hp_obj.upper),name=hp)
            elif 'float' in hp_type:
                if hp_obj.log: # log10-scale hyperparmeter
                    hp_dimension = Real(low=float(hp_obj.lower), high=float(hp_obj.upper), prior='log-uniform', name=hp)
                else:
                    hp_dimension = Real(low=float(hp_obj.lower), high=float(hp_obj.upper), name=hp)
            elif 'categorical' in hp_type:
                hp_dimension = Categorical(hp_obj.choices, name=hp)
            elif 'ordinal' in hp_type:
                hp_dimension = Categorical(hp_obj.sequence, name=hp)
            else:
                raise ValueError("unknown hyperparameter type: %s" % hp)
            skopt_hpspace.append(hp_dimension)
        skopt_keys = {
            'base_estimator', 'n_random_starts', 'n_initial_points',
            'acq_func', 'acq_optimizer', 'random_state',  'model_queue_size',
            'acq_func_kwargs', 'acq_optimizer_kwargs'}
        skopt_kwargs = self._filter_skopt_kwargs(kwargs, skopt_keys)
        self.bayes_optimizer = Optimizer(
            dimensions=skopt_hpspace, **skopt_kwargs)

    @staticmethod
    def _filter_skopt_kwargs(kwargs, keys):
        return {k: v for k, v in kwargs.items() if k in keys}

    def configure_scheduler(self, scheduler):
        from ..scheduler.fifo import FIFOScheduler

        assert isinstance(scheduler, FIFOScheduler), \
            "This searcher requires FIFOScheduler scheduler"
        super().configure_scheduler(scheduler)

    def get_config(self, **kwargs):
        """Function to sample a new configuration
        This function is called to query a new configuration that has not yet been tried.
        Asks for one point at a time from skopt, up to max_tries. 
        If an invalid hyperparameter configuration is proposed by skopt, then reverts to random search
        (since skopt configurations cannot handle conditional spaces like ConfigSpace can).
        TODO: may loop indefinitely due to no termination condition (like RandomSearcher.get_config() ) 
        
        Parameters
        ----------
        max_tries: int, default = 1e2
            The maximum number of tries to ask for a unique config from skopt before reverting to random search.
        """
        max_tries = kwargs.get('max_tries', 1e2)
        if len(self._results) == 0: # no hyperparams have been tried yet, first try default config
            return self.default_config()
        try:
            new_points = self.bayes_optimizer.ask(n_points=1) # initially ask for one new config
            new_config_cs = self.skopt2config(new_points[0]) # hyperparameter-config to evaluate
            try:
                new_config_cs.is_valid_configuration()
                new_config = new_config_cs.get_dictionary()
                if pickle.dumps(new_config) not in self._results.keys(): # have not encountered this config
                    self._results[pickle.dumps(new_config)] = self._reward_while_pending()
                    return new_config
            except self.errors_tohandle:
                pass
            new_points = self.bayes_optimizer.ask(n_points=max_tries) # ask skopt for many configs since first one was not new
            i = 1 # which new point to return as new_config, we already tried the first point above
            while i < max_tries:
                new_config_cs = self.skopt2config(new_points[i]) # hyperparameter-config to evaluate
                try:
                    new_config_cs.is_valid_configuration()
                    new_config = new_config_cs.get_dictionary()
                    if (pickle.dumps(new_config) not in self._results.keys()): # have not encountered this config
                        self._results[pickle.dumps(new_config)] = self._reward_while_pending()
                        return new_config
                except self.errors_tohandle:
                    pass
                i += 1
        except self.errors_tohandle:
            pass
        logger.info("used random search instead of skopt to produce new hyperparameter configuration in this trial")
        return self.random_config()
    
    def default_config(self):
        """ Function to return the default configuration that should be tried first.
        
        Returns
        -------
        returns: config
        """
        new_config_cs = self.configspace.get_default_configuration()
        new_config = new_config_cs.get_dictionary()
        self._results[pickle.dumps(new_config)] = self._reward_while_pending()
        return new_config
        
    def random_config(self):
        """Function to randomly sample a new configuration (which is ensured to be valid in the case of conditional hyperparameter spaces).
        """
        # TODO: may loop indefinitely due to no termination condition (like RandomSearcher.get_config() ) 
        new_config = self.configspace.sample_configuration().get_dictionary()
        while pickle.dumps(new_config) in self._results.keys():
            new_config = self.configspace.sample_configuration().get_dictionary()
        self._results[pickle.dumps(new_config)] = self._reward_while_pending()
        return new_config

    def update(self, config, **kwargs):
        """Update the searcher with the newest metric report.
        """
        super().update(config, **kwargs)
        reward = kwargs[self._reward_attribute]
        try:
            self.bayes_optimizer.tell(self.config2skopt(config),
                                      -reward)  # provide negative reward since skopt performs minimization
        except self.errors_tohandle:
            logger.info("surrogate model not updated this trial")

    def config2skopt(self, config):
        """ Converts autogluon config (dict object) to skopt format (list object).

        Returns
        -------
        Object of same type as: `skOpt.Optimizer.ask()`
        """
        point = []
        for hp in self.hp_ordering:
            point.append(config[hp])
        return point
    
    def skopt2config(self, point):
        """ Converts skopt point (list object) to autogluon config format (dict object. 
        
        Returns
        -------
        Object of same type as: `RandomSampling.configspace.sample_configuration().get_dictionary()`
        """
        config = self.configspace.sample_configuration()
        for i in range(len(point)):
            hp = self.hp_ordering[i]
            config[hp] = point[i]
        return config



import logging
import math
import pickle
import sys
import time

import numpy as np
import psutil
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor

from ..abstract import model_trial
from ..abstract.abstract_model import AbstractModel
from ...constants import MULTICLASS, REGRESSION
from ....features.generators import LabelEncoderFeatureGenerator
from ....utils.exceptions import NotEnoughMemoryError, TimeLimitExceeded

logger = logging.getLogger(__name__)


class RFModel(AbstractModel):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._model_type = self._get_model_type()
        self._feature_generator = None

    def _get_model_type(self):
        if self.problem_type == REGRESSION:
            return RandomForestRegressor
        else:
            return RandomForestClassifier

    # TODO: X.fillna -inf? Add extra is_missing column?
    def preprocess(self, X):
        X = super().preprocess(X)
        if self._feature_generator is None:
            self._feature_generator = LabelEncoderFeatureGenerator(verbosity=0)
            self._feature_generator.fit(X=X)
        if self._feature_generator.features_in:
            X = X.copy()
            X[self._feature_generator.features_in] = self._feature_generator.transform(X=X)
        X = X.fillna(0).to_numpy(dtype=np.float32)
        return X

    def _set_default_params(self):
        default_params = {
            'n_estimators': 300,
            'n_jobs': -1,
            'random_state': 0,
        }
        for param, val in default_params.items():
            self._set_default_param_value(param, val)

    # TODO: Add in documentation that Categorical default is the first index
    # TODO: enable HPO for RF models
    def _get_default_searchspace(self):
        spaces = {
            # 'n_estimators': Int(lower=10, upper=1000, default=300),
            # 'max_features': Categorical(['auto', 0.5, 0.25]),
            # 'criterion': Categorical(['gini', 'entropy']),
        }
        return spaces

    def _fit(self, X_train, y_train, time_limit=None, **kwargs):
        time_start = time.time()
        max_memory_usage_ratio = self.params_aux['max_memory_usage_ratio']
        hyperparams = self.params.copy()
        n_estimators_final = hyperparams['n_estimators']

        n_estimators_minimum = min(40, n_estimators_final)
        if n_estimators_minimum < 40:
            n_estimators_test = max(1, math.floor(n_estimators_minimum/5))
        else:
            n_estimators_test = 8

        X_train = self.preprocess(X_train)
        n_estimator_increments = [n_estimators_final]

        # Very rough guess to size of a single tree before training
        if self.problem_type == MULTICLASS:
            if self.num_classes is None:
                num_trees_per_estimator = 10  # Guess since it wasn't passed in, could also check y_train for a better value
            else:
                num_trees_per_estimator = self.num_classes
        else:
            num_trees_per_estimator = 1
        bytes_per_estimator = num_trees_per_estimator * len(X_train) / 60000 * 1e6  # Underestimates by 3x on ExtraTrees
        available_mem = psutil.virtual_memory().available
        expected_memory_usage = bytes_per_estimator * n_estimators_final / available_mem
        expected_min_memory_usage = bytes_per_estimator * n_estimators_minimum / available_mem
        if expected_min_memory_usage > (0.5 * max_memory_usage_ratio):  # if minimum estimated size is greater than 50% memory
            logger.warning(f'\tWarning: Model is expected to require {round(expected_min_memory_usage * 100, 2)}% of available memory (Estimated before training)...')
            raise NotEnoughMemoryError

        if n_estimators_final > n_estimators_test * 2:
            if self.problem_type == MULTICLASS:
                n_estimator_increments = [n_estimators_test, n_estimators_final]
                hyperparams['warm_start'] = True
            else:
                if expected_memory_usage > (0.05 * max_memory_usage_ratio):  # Somewhat arbitrary, consider finding a better value, should it scale by cores?
                    # Causes ~10% training slowdown, so try to avoid if memory is not an issue
                    n_estimator_increments = [n_estimators_test, n_estimators_final]
                    hyperparams['warm_start'] = True

        hyperparams['n_estimators'] = n_estimator_increments[0]
        self.model = self._model_type(**hyperparams)

        time_train_start = time.time()
        for i, n_estimators in enumerate(n_estimator_increments):
            if i != 0:
                self.model.n_estimators = n_estimators
            self.model = self.model.fit(X_train, y_train)
            if (i == 0) and (len(n_estimator_increments) > 1):
                time_elapsed = time.time() - time_train_start

                model_size_bytes = sys.getsizeof(pickle.dumps(self.model))
                expected_final_model_size_bytes = model_size_bytes * (n_estimators_final / self.model.n_estimators)
                available_mem = psutil.virtual_memory().available
                model_memory_ratio = expected_final_model_size_bytes / available_mem

                ideal_memory_ratio = 0.15 * max_memory_usage_ratio
                n_estimators_ideal = min(n_estimators_final, math.floor(ideal_memory_ratio / model_memory_ratio * n_estimators_final))

                if n_estimators_final > n_estimators_ideal:
                    if n_estimators_ideal < n_estimators_minimum:
                        logger.warning(f'\tWarning: Model is expected to require {round(model_memory_ratio*100, 2)}% of available memory...')
                        raise NotEnoughMemoryError  # don't train full model to avoid OOM error
                    logger.warning(f'\tWarning: Reducing model \'n_estimators\' from {n_estimators_final} -> {n_estimators_ideal} due to low memory. Expected memory usage reduced from {round(model_memory_ratio*100, 2)}% -> {round(ideal_memory_ratio*100, 2)}% of available memory...')

                if time_limit is not None:
                    time_expected = time_train_start - time_start + (time_elapsed * n_estimators_ideal / n_estimators)
                    n_estimators_time = math.floor((time_limit - time_train_start + time_start) * n_estimators / time_elapsed)
                    if n_estimators_time < n_estimators_ideal:
                        if n_estimators_time < n_estimators_minimum:
                            logger.warning(f'\tWarning: Model is expected to require {round(time_expected, 1)}s to train, which exceeds the maximum time limit of {round(time_limit, 1)}s, skipping model...')
                            raise TimeLimitExceeded
                        logger.warning(f'\tWarning: Reducing model \'n_estimators\' from {n_estimators_ideal} -> {n_estimators_time} due to low time. Expected time usage reduced from {round(time_expected, 1)}s -> {round(time_limit, 1)}s...')
                        n_estimators_ideal = n_estimators_time

                for j in range(len(n_estimator_increments)):
                    if n_estimator_increments[j] > n_estimators_ideal:
                        n_estimator_increments[j] = n_estimators_ideal

        self.params_trained['n_estimators'] = self.model.n_estimators

    def hyperparameter_tune(self, X_train, y_train, X_val, y_val, scheduler_options=None, **kwargs):
        fit_model_args = dict(X_train=X_train, y_train=y_train, **kwargs)
        predict_proba_args = dict(X=X_val)
        model_trial.fit_and_save_model(model=self, params=dict(), fit_args=fit_model_args, predict_proba_args=predict_proba_args, y_val=y_val, time_start=time.time(), time_limit=None)
        hpo_results = {'total_time': self.fit_time}
        hpo_model_performances = {self.name: self.val_score}
        hpo_models = {self.name: self.path}
        return hpo_models, hpo_model_performances, hpo_results

    def get_model_feature_importance(self):
        if self.features is None:
            # TODO: Consider making this raise an exception
            logger.warning('Warning: get_model_feature_importance called when self.features is None!')
            return dict()
        return dict(zip(self.features, self.model.feature_importances_))



import os
import numpy as np
from sklearn.impute import SimpleImputer
from sklearn.model_selection import train_test_split

from .files import download

__all__ = ['load_and_split_openml_data']


def load_and_split_openml_data(
        openml_task_id=6, ratio_train_valid=0.33, download_from_openml=False):
    """Download OpenML dataset (task), split of training set

    :param openml_task_id:
    :param ratio_train_valid:
    :param download_from_openml: If False, the data is downloaded from the
        AutoGluon S3 bucket. This works only for tasks which have been
        uploaded there. Use this if the code is run as part of the AutoGluon
        CI system (downloading from OpenML fails too often)
    :return: X_train, X_valid, y_train, y_valid, n_classes
    """
    import openml  # Dependence for tests and docs only
    if not download_from_openml:
        # We download the data from the AutoGluon S3 bucket, avoiding a
        # download from OpenML, which fails too often
        src_url = 'https://autogluon.s3.amazonaws.com/'
        trg_path = './'
        data_path = 'org/openml/www/datasets/{}/'.format(openml_task_id)
        task_path = 'org/openml/www/tasks/{}/'.format(openml_task_id)
        data_files = [
            data_path + x for x in [
                'dataset.arff', 'dataset.pkl.py3', 'description.xml',
                'features.xml', 'qualities.xml']] + [
            task_path + x for x in [
                'datasplits.arff', 'datasplits.pkl.py3', 'task.xml']]
        for fname in data_files:
            trg_fname = trg_path + fname
            if not os.path.exists(trg_fname):
                download(src_url + fname, path=trg_fname)
    # Note: If the files already exist locally, openml will not download them
    openml.config.set_cache_directory("./")
    task = openml.tasks.get_task(openml_task_id)
    n_classes = len(task.class_labels)
    train_indices, test_indices = task.get_train_test_split_indices()
    X, y = task.get_X_and_y()
    # Basic missing values imputation
    imputer = SimpleImputer(missing_values=np.nan, strategy='mean')
    X = imputer.fit_transform(X)
    X_train = X[train_indices]
    y_train = y[train_indices]
    # Train/validation split and standardization of inputs
    X_train, X_valid, y_train, y_valid = train_test_split(
        X_train, y_train, random_state=1, test_size=ratio_train_valid)
    mean = np.mean(X_train, axis=0)
    std = np.std(X_train, axis=0)
    X_train = (X_train - mean) / (std + 1e-10)
    X_valid = (X_valid - mean) / (std + 1e-10)

    return X_train, X_valid, y_train, y_valid, n_classes



from . import constants as _C
from autogluon_contrib_nlp.data import batchify as bf


class _TextTokenIdsFieldBatchify:
    def __init__(self, round_to=None):
        self._pad_batchify = bf.Pad(round_to=round_to)
        self._stack_batchify = bf.Stack()

    def __call__(self, data):
        """

        Parameters
        ----------
        data

        Returns
        -------
        batch_token_ids
            (batch_size, sequence_length)
        batch_valid_length
            (batch_size,)
        batch_segment_ids
            (batch_size, sequence_length)
        batch_token_offsets
            (batch_size, sequence_length, 2)
        """
        batch_token_ids = self._pad_batchify([ele.token_ids for ele in data])
        batch_valid_length = self._stack_batchify([len(ele.token_ids) for ele in data])
        if data[0].segment_ids is None:
            batch_segment_ids = None
        else:
            batch_segment_ids = self._pad_batchify([ele.segment_ids for ele in data])
        if data[0].token_offsets is None:
            batch_token_offsets = None
        else:
            batch_token_offsets = self._pad_batchify([ele.token_offsets for ele in data])
        return batch_token_ids, batch_valid_length, batch_segment_ids, batch_token_offsets


class TextTokenIdsField:
    type = _C.TEXT

    def __init__(self, token_ids, segment_ids=None, token_offsets=None):
        """

        Parameters
        ----------
        token_ids
            The token_ids, shape (seq_length,)
        segment_ids
            The segment_ids, shape (seq_length,)
        token_offsets
            The character-level offsets of the token, shape (seq_length, 2)
        """
        self.token_ids = token_ids
        self.segment_ids = segment_ids
        self.token_offsets = token_offsets

    @staticmethod
    def batchify(round_to=None):
        return _TextTokenIdsFieldBatchify(round_to)

    def __str__(self):
        ret = '{}(\n'.format(self.__class__.__name__)
        ret += 'token_ids={}\n'.format(self.token_ids)
        ret += 'segment_ids={}\n'.format(self.segment_ids)
        ret += 'token_offsets={}\n'.format(self.token_offsets)
        ret += ')\n'
        return ret


class _EntityFieldBatchify:
    def __init__(self):
        self._pad_batchify = bf.Pad()
        self._stack_batchify = bf.Stack()

    def __call__(self, data):
        """The internal batchify function

        Parameters
        ----------
        data
            The input data.

        Returns
        -------
        batch_span
            Shape (batch_size, #num_entities, 2)
        batch_label
            Shape (batch_size, #num_entities) + label_shape
        batch_num_entity
            Shape (batch_size,)
        """
        batch_span = self._pad_batchify([ele.data for ele in data])
        no_label = data[0].label is None
        if no_label:
            batch_label = None
        else:
            batch_label = self._pad_batchify([ele.label for ele in data])
        batch_num_entity = self._stack_batchify([len(ele.data) for ele in data])
        return batch_span, batch_label, batch_num_entity


class EntityField:
    type = _C.ENTITY

    def __init__(self, data, label=None):
        """

        Parameters
        ----------
        data
            (#Num Entities, 2)
        label
            (#Num Entities,)
        """
        self.data = data
        self.label = label

    @staticmethod
    def batchify():
        return _EntityFieldBatchify()

    def __str__(self):
        ret = '{}(\n'.format(self.__class__.__name__)
        ret += 'data={}\n'.format(self.data)
        ret += 'label={}\n'.format(None if self.label is None else self.label)
        ret += ')\n'
        return ret


class _ArrayBatchify:
    def __init__(self):
        self._stack_batchify = bf.Stack()

    def __call__(self, data):
        """

        Parameters
        ----------
        data

        Returns
        -------
        dat
            Shape (batch_size,) + sample_shape
        """
        return self._stack_batchify([ele.data for ele in data])


class NumericalField:
    type = _C.NUMERICAL

    def __init__(self, data):
        self.data = data

    @staticmethod
    def batchify():
        return _ArrayBatchify()

    def __str__(self):
        ret = '{}(\n'.format(self.__class__.__name__)
        ret += 'data={}\n'.format(self.data)
        ret += ')\n'
        return ret


class CategoricalField(NumericalField):
    type = _C.CATEGORICAL
    pass
